-module(utLifo).

-export([
   new/1
   , del/1
   , in/2
   , out/1
   , clear/1
   , size/1
]).

-spec new(Name :: atom()) -> ok | name_used.
new(Name) ->
   case persistent_term:get(Name, undefined) of
      undefined ->
         ARef = atomics:new(1, [{signed, false}]),
         ets:new(Name, [set, named_table, {write_concurrency, true}]),
         persistent_term:put(Name, ARef),
         ok;
      _ ->
         case ets:info(Name, id) of
            undefined ->
               ets:new(Name, [set, named_table, {write_concurrency, true}]);
            _ ->
               name_used
         end
   end.

-spec del(Name :: atom()) -> ok | error_lifo.
del(Name) ->
   case persistent_term:get(Name, undefined) of
      undefined ->
         error_lifo;
      _ ->
         persistent_term:erase(Name),
         ets:delete(Name),
         ok
   end.

-spec in(Name :: atom(), Value :: term()) -> true | false | error_lifo.
in(Name, Value) ->
   case persistent_term:get(Name, undefined) of
      undefined ->
         error_lifo;
      ARef ->
         in(ARef, Name, Value)
   end.

in(ARef, Name, Value) ->
   Index = atomics:add_get(ARef, 1, 1),
   case ets:insert_new(Name, {Index, Value}) of
      true ->
         true;
      _ ->
         in(ARef, Name, Value)
   end.

-spec out(Name :: atom()) -> error_lifo | empty | Value :: term().
out(Name) ->
   case persistent_term:get(Name, undefined) of
      undefined ->
         error_lifo;
      ARef ->
         out(ARef, Name)
   end.

out(ARef, Name) ->
   Index = atomics:sub_get(ARef, 1, 1),
   case ets:take(Name, Index + 1) of
      [] ->
         case ets:info(Name, size) of
            0 ->
               empty;
            _ ->
               out(ARef, Name)
         end;
      [{_, Value}] ->
         Value
   end.

-spec clear(Name :: atom()) -> error_lifo | ok.
clear(Name) ->
   case persistent_term:get(Name, undefined) of
      undefined ->
         error_lifo;
      _ ->
         ets:delete_all_objects(Name),
         ok
   end.

-spec size(Name :: atom()) -> Size :: integer() | error_lifo.
size(Name) ->
   case ets:info(Name, size) of
      undefined ->
         error_lifo;
      Size ->
         Size
   end.

